<div
  {% if widget._center_content == true %}
  style="display: flex; gap: {{{widget._gap}}}px; justify-content: center;"
  {% else %}
  style="display: flex; gap: {{{widget._gap}}}px;"
  {% endif %}
>
  {% for w in widget._widgets %}
    {{{w}}}
  {% endfor %}
</div>
